#include <thread>
#include <s7.h>
#include <mutex>
#include <thread>
#include "stdio.h"
std::recursive_mutex mutex;

#define HELP_set_timeout "(set-timeout timeout thunk)"
s7_pointer scm_set_timeout(s7_scheme* sc, s7_pointer args) {
  s7_pointer timeout = s7_car(args); args = s7_cdr(args);
  s7_pointer thunk = s7_car(args); args = s7_cdr(args);
  int i_timeout = s7_integer(timeout);
  int gc = s7_gc_protect(sc, thunk);
  std::thread t([=](){
    // thread will call this c++ "thunk", which has captured used variables
    std::this_thread::sleep_for(std::chrono::milliseconds(i_timeout));
    // locking RAII style: gets unlocked when it goes out of scope
    std::unique_lock<std::recursive_mutex> lock(mutex);
    s7_call(sc, thunk, s7_nil(sc));
    s7_gc_unprotect_at(sc, gc);
    });
  // letting the thread run in the background, while return s7_nil
  t.detach();

  return s7_nil(sc);
}
void bind_set_timeout(s7_scheme* sc, s7_pointer env) {
  s7_define(sc, env, s7_make_symbol(sc, "set-timeout"),
            s7_make_function(sc, "set-timeout",
                             scm_set_timeout,
                             2, 0, false, // req, opt, rest?
                             HELP_set_timeout)
            );
}


int main(int argc, char** argv) {
  printf("hi from c\n");
  s7_scheme* sc = s7_init();
  bind_set_timeout(sc, s7_nil(sc));
  std::unique_lock<std::recursive_mutex> lock(mutex, std::defer_lock);
  // locking, just in case any (set-timeout ..) is called from main.scm
  lock.lock();
  s7_load(sc, "main.scm");
  lock.unlock();
  char buffer[512];
  bool running = true;
  printf("\n> ");
  while(running) {
    if (fgets(buffer, 512, stdin) == NULL) {
      running = false;
      break;
    }
    s7_pointer res = s7_eval_c_string(sc, buffer);
    // dummy repl
    printf("%s\n> ", s7_string(s7_object_to_string(sc, res, false)));
  }
  s7_free(sc);

  return 0;
}
